#!/usr/bin/env python
# -*- coding: utf-8 -*-

##
# Copyright (C) 2009,2010 Mathieu Stute, http://wookr.com/pydaemon
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##

"""
==============================
pydaemon -- Python Unix daemon
==============================

This module allow you to easily turn your Python scripts to Unix daemons. You
just have to create a class that inherits from :class:`Daemon` and to put your
code to the :meth:`run` method.

Dependency: RSFile (http://pypi.python.org/pypi/RSFile/)

Here is a simple and useless example that shows how to use :mod:`pydaemon`
module: ::

  import logging
  import time

  from pydaemon import Daemon

  class FooDaemon(Daemon):
    def run(self):
      while True:
        logging.debug("I'm here...")
        time.sleep(1)

    def bar(self):
      logging.debug("bar")

  if __name__ == "__main__":
    logging.basicConfig(filename="foodaemon.log",level=logging.DEBUG)
    daemon = FooDaemon("/tmp/foodaemon.pid", "FooDaemon")
    daemon.main(extended_args={"bar": daemon.bar})

"""

import atexit
import os
import signal
import sys
import time

from rsfile import rsopen

__version__ = '0.2.3'
__author__ = 'Mathieu Stute'

class Daemon:
  """
  Unix daemon class that does nothing on its own. You need to subclass it and
  override the run() method.
  """
  def __init__(self, pidfile, name="", working_dir="/", stdin='/dev/null',
               stdout='/dev/null', stderr='/dev/null', uid=None):
    if isinstance(uid, int):
      try:
        os.setuid(uid)
      except:
        sys.stderr.write("cannot set uid %s for the daemon\n" % uid)
        sys.exit(4)
    self.stdin = stdin
    self.stdout = stdout
    self.stderr = stderr
    self.pidfile = pidfile
    self.lock = None
    self.working_dir = working_dir
    self.daemon_name = name
    signal.signal(signal.SIGTERM, self.__signals_handler)

  def __signals_handler(self, signum, current_stack_frame):
    """
    Intercepts SIGTERM to call cleanup() and to cleanly exit.
    """
    if signum == signal.SIGTERM:
      if self.cleanup() is None:
        os._exit(0)

  def __daemonize(self):
    """
    Does the UNIX double-fork magic, see Stevens' "Advanced Programming in the
    UNIX Environment" for details (ISBN 0201563177)
    http://www.yendor.com/programming/unix/apue/proc/fork2.c
    """
    def delpid():
      os.remove(self.pidfile)

    try:
      # first fork
      pid = os.fork()
      if pid > 0:
        # exit from first parent
        os._exit(0)
    except OSError, e:
      sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
      os._exit(1)

    # decouple from parent environment
    os.chdir(self.working_dir)
    os.setsid()
    os.umask(0)
    # second fork
    try:
      pid = os.fork()
      if pid > 0:
        # exit from second parent
        os._exit(0)
    except OSError, e:
      sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
      sys.exit(1)

    # redirects standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    si = file(self.stdin, 'r')
    so = file(self.stdout, 'a+')
    se = file(self.stderr, 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    # writes the pidfile, lock it and register delpid() on program exit.
    atexit.register(delpid)
    pid = str(os.getpid())
    file(self.pidfile,'w+').write("%s\n" % pid)
    self.__lock()

  def __lock(self):
    """
    Tries to lock the pidfile. Returns True if the lock succeed otherwise False.
    """
    try:
      self.lock = rsopen(self.pidfile, locking=True, mode="W", timeout=0.1)
      return True
    except:
      return False

  def _status(self, exit=False):
    """
    Checks if the daemon is running. Warning: when exit is set to True sys.exit
    will be called.
    """
    if not self.__lock():
      if exit:
        sys.stdout.write('%s is running\n' % self.daemon_name)
        sys.exit(1)
      return True
    else:
      if exit:
        sys.stdout.write('%s is not running\n' % self.daemon_name)
        sys.exit(0)
      del self.lock
      return False

  def _start(self):
    """
    Starts the daemon.
    """
    if not self.__lock():
      sys.stderr.write("%s is already running\n" % (self.daemon_name))
      sys.exit(1)
    self.setup()
    sys.stdout.write('starting %s\n' % self.daemon_name)
    self.__daemonize()
    self.run()

  def _stop(self):
    """
    Stops the daemon.
    """

    # Tries to read the process id from the pid file
    try:
      pf = file(self.pidfile,'r')
      pid = int(pf.read().strip())
      pf.close()
    except ValueError:
      sys.stderr.write("failed to read pid file\n")
      sys.exit(2)
    except IOError:
      sys.stderr.write("%s is not running.\n" % self.daemon_name)
      sys.exit(1)

    # Tries to kill the daemon process
    sys.stdout.write('stoping %s ...\n' % self.daemon_name)
    try:
      while 1:
        os.kill(pid, signal.SIGTERM)
        time.sleep(0.1)
    except OSError, err:
      err = str(err)
      # normal exit, the process has been killed, os.kill raises an OSError.
      if err.find("No such process") > 0:
        if os.path.exists(self.pidfile):
          os.remove(self.pidfile)
      else:
        print str(err)
        sys.exit(3)

  def _restart(self):
    """
    Restarts the daemon.
    """
    if self._status():
      self._stop()
    self._start()

  def setup(self):
    """
    Called just before the code is daemonized.
    """

  def run(self):
    """
    Override this method when you subclass Daemon. It will be called after the
    process has been daemonized by start() or restart(). Just put your main loop
    here.
    """

  def cleanup(self):
    """
    Implement this method to perform cleanup actions before being stopped.
    If the returned value is not None the daemon will not be stopped. You can
    cleanly stop the daemon by calling _stop().
    """

  def _help(self, extended_args):
    """
    Prints a help message to stdout.
    """
    m = "%s accepts these commands:\n  start\n  stop\n  restart\n  status\n"\
         % (sys.argv[0])
    for arg in extended_args:
      m += "  %s\n" % (arg)
    sys.stdout.write(m)

  def main(self, extended_args={}):
    """
    Use this method in your main program.
    You can extend the default valid arguments (start, stop, status, restart) by
    using extended_args. It is a dict where keywords are the commands and values
    are the associated functions.
    """
    cmd = sys.argv[len(sys.argv) - 1]
    if cmd == 'start':
      self._start()
    elif cmd == 'stop':
      self._stop()
    elif cmd == 'restart':
      self._restart()
    elif cmd == 'status':
      self._status(True)
    elif cmd in ["--help", "-h", "help"]:
      self._help(extended_args)
    else:
      if extended_args.has_key(cmd):
        try:
          fn = extended_args[cmd]
          fn()
        except Exception, e:
          sys.stderr.write("pydaemon: \"%s\" argument error: %s\n" % (cmd, e))
      else:
        if cmd != sys.argv[0]:
          sys.stderr.write("unknown command: %s\n" % (cmd))
        self._help(extended_args)
        sys.exit(2)
    sys.exit(0)
